            // A simple UDP file transfer client.
            // The protocol is modelled on the Trivial File Transfer
            // Protocol (TFTP) described in RFC 1350.
            // Only receipt of a file is implemented, and the
            // transfer mode is binary (i.e. octet).

import java.net.*;
import java.io.*;

class UDPFileClient extends UDPUser {
        // Contact the UDP file server running on hostname.
    public UDPFileClient(String hostname) throws UnknownHostException,
                                                 SocketException {
        super(hostname);
    }

        // Retrieve the given file from host.
    public void run(String filename) throws Exception {
        // The local copy of the file.
        OutputStream outStream = null;
        try{
            // Create a buffer and packet for receiving replies.
            // The data is sent in fixed sized packets so we can
            // use the same packet for all replies.
            byte[] replySpace = new byte[UDPFileTransfer.fullDataPacket];
            DatagramPacket reply =
                new DatagramPacket(replySpace,replySpace.length);

            // Open the local copy of the file.
            outStream = new FileOutputStream(filename);

            // Try to connect, and get the port of the server's slave
            // thread. The reply will contain the first block
            // of data.
            int slavePort = connect(filename,reply);

            // Create space for an ACK packet.
            byte[] ack = new byte[UDPFileTransfer.opcodeSize+
                                  UDPFileTransfer.blockNumberSize];
            DatagramPacket ackPacket = new DatagramPacket(ack,ack.length,
                                           getCorrespondent(),slavePort);
            // Encode the ACK opcode.
            twoByteEncode(UDPFileTransfer.ACK,ack,0);
            retrieve(reply,ackPacket,outStream);
        }
        catch(IOException e){
            System.err.println("IOException: "+e.getMessage());
        }
        finally{
            if(outStream != null){
                outStream.close();
            }
            getSocket().close();
        }
    }

        // Make contact with the server. If successful, it should return
        // the number of the server port to use in all subsequent
        // communication.
    protected int connect(String filename,DatagramPacket reply)
                                        throws IOException {
        // Get the socket we are using.
        DatagramSocket socket = getSocket();
        // Get the address of the server.
        InetAddress server = getCorrespondent();
        // Turn the name of the file into a byte array; this is the
        // form it must be sent in the initial request.
        byte[] filenameBytes = filename.getBytes();

        // Create a byte array to contain the read request.
        // The format is RRQ+filename+0.
        byte[] request = new byte[UDPFileTransfer.opcodeSize+
                                  filenameBytes.length+1];
        // Store the opcode.
        twoByteEncode(UDPFileTransfer.RRQ,request,0);
        // Copy in the name after the opcode.
        System.arraycopy(filenameBytes,0,request,UDPFileTransfer.opcodeSize,
                                filenameBytes.length);
        // Terminate the name.
        request[request.length-1] = 0;
        // Create the packet to send the request.
        DatagramPacket packet =
                    new DatagramPacket(request,request.length,
                    server,UDPFileTransfer.ServerPort);
        sendAndReceive(packet,reply);

        // The first reply should contain the first DATA packet.
        byte[] data = reply.getData();
        if(lengthOk(reply.getLength()) && opcodeOk(data) &&
                                        expectedBlock(data,1)){
            // Return the slave's port number.
            return reply.getPort();
        }
        else{
            throw new IOException("Failed to receive the first data packet.");
        }
    }

        // Perform the main retrieval. The first DATA block is
        // already in packet and is still to be acknowledged.
        // Write the data to outStream.
    protected void retrieve(DatagramPacket packet,DatagramPacket ackPacket,
                            OutputStream outStream)
                            throws Exception {
        // Keep track of the next block we expect to receive.
        int nextBlockNumber = 1;

        boolean endOfTransfer = false;
        // Each time around the loop check that the last packet
        // received is valid and contains the block we expect.
        while(!endOfTransfer){
            // Access the data space and the length of the packet.
            byte[] data = packet.getData();
            int length = packet.getLength();

            if(!lengthOk(length)){
                throw new Exception("Bad length packet received.");
            }
            else if(!opcodeOk(data)){
                throw new Exception("Non-data packet received.");
            }
            else if(expectedBlock(data,nextBlockNumber)){
                System.out.println("Received block: "+nextBlockNumber);
                // Write it out (omitting the header).
                outStream.write(data,UDPFileTransfer.dataHeaderLength,
                            length-UDPFileTransfer.dataHeaderLength);

                // Encode the block number in the acknowledgement.
                twoByteEncode(nextBlockNumber,ackPacket.getData(),
                              UDPFileTransfer.opcodeSize);

                // Check to see whether this was the last block.
                if(length == UDPFileTransfer.fullDataPacket){
                    // It wasn't, so we expect another.
                    nextBlockNumber++;
                    // Receive the next packet.
                    sendAndReceive(ackPacket,packet);
                }
                else{
                    System.out.println("Final block received.");
                    // Acknowledge it.
                    send(ackPacket);
                    endOfTransfer = true;
                }
            }
            else{
                // Resend the last acknowledgement.
                sendAndReceive(ackPacket,packet);
            }
        }
    }

        // Make sure the received data is long enough to
        // contain at least an opcode and block number.
    protected boolean lengthOk(int length){
        return (length >= UDPFileTransfer.opcodeSize+
                         UDPFileTransfer.blockNumberSize) &&
               (length <= UDPFileTransfer.fullDataPacket);
    }

        // Make sure the received data is tagged as DATA.
    protected boolean opcodeOk(byte[] data){
        return (data[0] == 0) && (data[1] == UDPFileTransfer.DATA);
    }

        // Make sure the received data is the right block number.
    protected boolean expectedBlock(byte[] data, int expected){
        int offset = UDPFileTransfer.opcodeSize;
        byte highByte = (byte) ((expected >> 8) & 0x0f);
        byte lowByte = (byte) (expected & 0x0f);
        return (data[offset] == highByte) && (data[offset+1] == lowByte);
    }

        // Store the lower order two bytes of value into data.
    protected void twoByteEncode(int value,byte[] data,int where){
        data[where] = (byte) ((value >> 8) & 0x0f);
        data[where+1] = (byte) (value & 0x0f);
    }
}
