            // Listener for a simple UDP file transfer server.
            // The protocol is modelled on the Trivial File Transfer
            // Protocol (TFTP) described in RFC 1350.
            // Only sending of a file is implemented, and the
            // transfer mode is binary (i.e. octet).

import java.net.*;
import java.io.*;

        // The part of the server that listens for connections.
        // As soon as a valid read request (RRQ) is received,
        // responsibility is handed over to a UDPFileTransferSlave run
        // as a separate thread.
class UDPFileServer {
        // Request a socket on the desired port.
    public UDPFileServer() throws SocketException {
        // Request to use a particular port.
        socket = new DatagramSocket(UDPFileTransfer.ServerPort);
    }

        // Listen for RRQ packets.
    public void run(){
        // Allocate space and a packet for receiving RRQ requests.
        byte[] data = new byte[UDPFileTransfer.fullDataPacket];
        DatagramPacket request = new DatagramPacket(data,data.length);
        DatagramSocket socket = getSocket();

        // Timeout after a period of inactivity.
        boolean keepGoing = true;
        while(keepGoing){
            // Wait for transfer requests to come in.
            try{
                // Give up after an hour (in milliseconds).
                final int hour = 60*60*1000;
                socket.setSoTimeout(hour);

                System.out.println("Waiting for requests.");
                socket.receive(request);

                System.out.println("Request received.");
                // Make sure the request is valid by checking some
                // basic packet format conditions.
                int length = request.getLength();

                if(!lengthOk(length)){
                    System.err.println("Illegal request length: "+length);
                }
                else if(!isReadRequest(data)){
                    System.err.println("Non RRQ packet received.");
                }
                else{
                    // Hand over to a slave.
                    new Thread(new UDPFileTransferSlave(request)).run();
                }
            }
            catch(SocketException e){
                System.err.println("Failed to create a slave thread.");
            }
            catch(InterruptedIOException e){
                keepGoing = false;
            }
            catch(IOException e){
                keepGoing = false;
            }
        }
        socket.close();
    }

        // Is the length of the request ok?
    protected boolean lengthOk(int length){
        // Allow at least one byte for a filename and a 0 terminator.
        return length >= (UDPFileTransfer.opcodeSize+1+1);
    }

    protected boolean isReadRequest(byte[] data){
        return (data[0] == 0) && (data[1] == UDPFileTransfer.RRQ);
    }

    protected DatagramSocket getSocket(){
        return socket;
    }

    private final DatagramSocket socket;
}
